#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import socket
import sys
import subprocess
import getopt
import logging
import time
import signal
import re
import json
from oslo_utils import netutils
import logging.handlers
from enum import IntEnum
from gi.repository import GLib
from pydbus import SystemBus
from threading import Thread, Event
# from scapy import all as sc
# from sysrepo_interfaces import sysrepo_interfaces3 as sr_if
# import default_gw_set as defgw

monitor_loop = True

LOG_NAME='tracehost' #DI: trace_host
OVPN_DEF_GW = '/tmp/ovpn_def_gw'
COLLECTD_PATH = '/tmp/collectd_sock'


class PR_TYPE(IntEnum):
    ICMP=1
    THROUGHPUT=2
    TCP=3
    HTTP=4
    HTTPS=5
    ARP=6
    DNS=7
    
class PR_THR_DIR(IntEnum):
    UPLOAD=1
    DOWNLOAD=2
    BOTH=3

class PR_ENDP_TYPE(IntEnum):
    IPV4=1
    IPV6=2
    HOSTNAME=3

class TracehostErr(Exception):
    def __init__(self, message, code):
        super(TracehostErr, self).__init__(message)
        self.code = code

class RttException(Exception):
    pass

class Tracehost(Thread):

    def __init__(self, ifname, if_brname=None, timeout=3, interval=10, count=3, ch_count=3, ifstate=True, trace_state=False, metric=None, state_prop=False,
                 pr_list={}, logger=None, log_level=logging.ERROR, is_vpp=False, ip_pass=False ):

        super(Tracehost, self).__init__()
        self.logger = None
        self.deamon = True
        self.cancelled = Event()
        
        if logger:
            self.logger = logger
        else:
            self.__init_logger( log_level ) 

        self._ifname = ifname
        if if_brname:
            self._br_name = if_brname
        else:
            self._br_name = 'br-' + self._ifname
        self._ifstate = ifstate
        self._ipv4 = False
        self._ipv6 = False

        # scapy
        self.ip4_addr = self.sc_ipv4_addr(self._br_name)
        if self.ip4_addr != None:
            self._ipv4 = True
            
        self.ip6_addr = self.sc_ipv6_addr(self._br_name)
        if self.ip6_addr != None:
            self._ipv6 = True

        self._trace_state = trace_state
        self._count = count
        self._ch_count = ch_count       
        self._interval = interval
        self._timeout = timeout
        self._metric = metric
        self._state_propagation = state_prop
        self._vpp_mode = is_vpp
        self._ippass = ip_pass
        
        # probes
        self._probes = pr_list
        for id, pr in self._probes.items():
            pr['tr_endpoint'] = None
            pr['tr_type'] = None
            self.check_endpoint_type(pr)
            if pr['type'] == PR_TYPE.THROUGHPUT:
                self.thr_set_server(pr)
        
        # Trace monitor variables
        self.need_change = False
        self.curr_state = True # active default GW
        
        # statistics
        self.upcount = 0
        self.downcount = 0

        # Probe id needed for collectd notifications
        self.probe_to_be_reported = None

    def __del__(self):
        self.logger.debug('__del__')
        
    def run(self):
        """ Overloaded Thread.run
            run probe by type        
        """

        while True:
            try:
                start = time.time()
                self.sampling()
                timediff = int(time.time()) - int(start)
            except TracehostErr as e:
                timediff = int(time.time()) - int(start)
                self.logger.error('run TracehostErr - exception {}' .format(e))
                # break
            except Exception as e:
                timediff = int(time.time()) - int(start)
                self.logger.error('run exception - {}' .format(e))
                # break

            # self.logger.debug("sampling time {}" .format(timediff))
            is_cancel = self.cancelled.wait(self._interval-timediff)
            if is_cancel:
                break
    
    def cancel(self):
        self.cancelled.set()
        for id, pr in self._probes.items():
            if pr['type'] == PR_TYPE.THROUGHPUT:
                self.set_th_ip_rules(pr['endpoint'], pr['tr_type'], is_del=True)
                self.flush_th_routes()
                self.set_th_vpp_ippass_rules(pr['endpoint'], pr['tr_type'], is_del=True)

    def __init_logger(self, log_level ):
        self.logger = logging.getLogger(name=LOG_NAME)
        self.logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self.logger.addHandler(syslog_hnd)
        self.logger.propagate = False

    @property
    def ifname(self):
        return self._ifname
    
    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, new):
        if new and not isinstance(new, bool):
            raise TracehostErr('Bad enable type param', -1)

        if self._enable == new:
            return

        self._enable = new
        # if self._enable == True:
        #     self.create_all_probes()
        # else:
        #     self.delete_all_probes()
        self.logger.info("Tracehost ifname {} enable - set new to {}".format(self._ifname, self._enable))

    def sc_ipv4_addr(self, ifname):
        ipaddr = sc.get_if_addr(ifname)
        if ipaddr != None and ipaddr != '0.0.0.0':
            return ipaddr
        return None

    def sc_ipv6_addr(self, ifname):
        return next((x[0] for x in sc.in6_getifaddr() if x[2] == ifname and x[1] == IPV6_ADDR_GLOBAL), None)

    def get_table_id(self):
        id = 40
        if 'data' in self._ifname:
            id = int(str(self._ifname).strip('data')) + 40
        elif 'lte' in self._ifname:
            id = int(str(self._ifname).strip('lte')) + 60
        elif 'wifi' in self._ifname:
            id = int(str(self._ifname).strip('wifi')) + 65
        elif 'ppp' in self._ifname:
            id = int(str(self._ifname).strip('ppp')) + 70
        return id

    def get_endpoint_ip(self, endpoint):
        end_ip = endpoint
        strings = ['http:', 'https:', '/']
        for s in strings:
            end_ip = str(endpoint).strip(str(s))
        
        # check if hostname
        return end_ip

    def check_endpoint_type(self, pr):
        if self._isIPv6(pr['endpoint']):
            pr['tr_type'] = PR_ENDP_TYPE.IPV6
        elif self._isIPv4(pr['endpoint']):
            pr['tr_type'] = PR_ENDP_TYPE.IPV4
        else:
            pr['tr_type'] = PR_ENDP_TYPE.HOSTNAME
        pr['tr_endpoint'] = self.get_endpoint_ip(pr['endpoint'])

    def thr_set_server(self, pr):
        FNULL = open(os.devnull, 'w')
        pr['srv_file'] = '/etc/speedtest_%s.json' % self._ifname
        try:
            subprocess.check_call(['/opt/bin/speedtest', '--set-server', str(pr['endpoint']), str(self._ifname)],
                                            stdout=FNULL, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.logger.error('Failed to set throughput server {} for interface {} with code {}' .format(pr['endpoint'], self._ifname, e.returncode))
        FNULL.close()
        
        pr['cmd'] = ["timeout", str(self._timeout), "librespeed-cli", "--local-json", pr['srv_file'], "--no-icmp", "--duration", "1", "--json", "--timeout", str(self._timeout)]
        if pr['tr_type'] == PR_ENDP_TYPE.IPV6:
            if not self.ip6_addr:
                self.logger.error('Configuration mismatch - interface {} with probe endpoint {} does not have an IPv6 address' 
                                  .format(self._br_name, pr['endpoint']))
            pr['cmd'].append('--source')
            pr['cmd'].append(str(self.ip6_addr))
        elif pr['tr_type'] == PR_ENDP_TYPE.IPV4:
            if not self.ip4_addr:
                self.logger.error('Configuration mismatch - interface {} with probe endpoint {} does not have an IPv4 address' 
                                  .format(self._br_name, pr['endpoint']))
            pr['cmd'].append('--source')
            pr['cmd'].append(str(self.ip4_addr))

        if pr['direction'] == PR_THR_DIR.UPLOAD:
            pr['cmd'].append('--no-download')
        if pr['direction'] == PR_THR_DIR.DOWNLOAD:
            pr['cmd'].append('--no-upload')

        self.set_th_routes()
        self.set_th_ip_rules(pr['endpoint'], pr['tr_type'], is_del=False)
        self.set_th_vpp_ippass_rules(pr['endpoint'], pr['tr_type'], is_del=False)

    def _isIPv6(self, ipaddr):
        """Check the IP address version and return True in case of IPv6"""
        if ipaddr is not None and ipaddr != "":
            return netutils.is_valid_ipv6(ipaddr)

        return False

    def _isIPv4(self, ipaddr):
        """Check the IP address version and return True in case of IPv4"""
        if ipaddr is not None and ipaddr != "":
            return netutils.is_valid_ipv4(ipaddr)

        return False

    def set_th_routes(self):
        sys_wan = self._br_name
        table_id = self.get_table_id()

        ipdb = IPDB()
        from pyroute2.netlink.exceptions import NetlinkError
        if sys_wan in ipdb.interfaces:
            wan_idx = ipdb.interfaces[sys_wan]
            for rt in ipdb.routes:
                # ARM has problem with iproute2, it returns IPv4 routes as IPV6 ones
                # if is_arm_arch and rt['family'] == 10:
                    # continue
                if rt['oif'] == wan_idx['index']:
                    if rt['gateway']:
                        if rt['gateway'].startswith("fe80:") : #link-local address
                            route = {'dst': rt['dst'], 'gateway': rt['gateway'], 'table': table_id, 'priority': rt['priority'], 'oif': wan_idx['index']}
                        else:
                            route = {'dst': rt['dst'], 'gateway': rt['gateway'], 'table': table_id, 'priority': rt['priority']}
                    else:
                        route = {'dst': rt['dst'], 'table': table_id, 'oif': wan_idx['index']}

                    try:
                        ipdb.routes.add(route).commit()
                    except NetlinkError as e:
                        self.logger.error('Set route in table {tbl} failed. {msg}'.format(tbl=table_id, msg=str(e)))
                        if e.code != 17:
                            self.logger.error('Set route in table {tbl} failed. {msg}'.format(tbl=table_id, msg=str(e)))

        ipdb.release()

    def flush_th_routes(self):
        FNULL = open(os.devnull, 'w')
        try:
            table_id = self.get_table_id()
            subprocess.check_call(['ip', 'route', 'flush' 'table', str(table_id)], stdout=FNULL, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            # self.logger.error('Failed to remove ip routes for interface {} with code {}' .format(self._ifname, e.returncode))
            pass
        FNULL.close()

    def set_th_ip_rules(self, endpoint, tr_type, is_del=False):
        FNULL = open(os.devnull, 'w')
        try:
            table_id = self.get_table_id()
            if tr_type == PR_ENDP_TYPE.HOSTNAME:
                FNULL.close()
                self.logger.error('Failed to configure ip rules. Endpoint {} is not a valid ip address' .format(endpoint))
                return
            else:
                end_ip = self.get_endpoint_ip(endpoint)
            
            if is_del:
                subprocess.check_call(['ip', 'rule', 'del', 'table', str(table_id)], stdout=FNULL, stderr=subprocess.STDOUT)
            else:
                if tr_type == PR_ENDP_TYPE.IPV6 and self.ip6_addr:
                    subprocess.check_call(['ip', 'rule', 'add', 'from', str(self.ip6_addr), 'to', str(end_ip), 
                                                    'table', str(table_id)], stdout=FNULL, stderr=subprocess.STDOUT)
                if tr_type == PR_ENDP_TYPE.IPV4 and self.ip4_addr:
                    subprocess.check_call(['ip', 'rule', 'add', 'from', str(self.ip4_addr), 'to', str(end_ip), 
                                                    'table', str(table_id)], stdout=FNULL, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.logger.error('Failed to configure ip rules for {} for interface {} with code {}' .format(endpoint, self._ifname, e.returncode))
            pass
        FNULL.close()

    def set_th_vpp_ippass_rules(self, endpoint, tr_type, is_del=False):
        if self._vpp_mode == False or self._ippass == False:
            return
        if tr_type == PR_ENDP_TYPE.HOSTNAME:
            self.logger.error('Failed to configure ip-pass rules. Endpoint {} is not a valid ip address' .format(endpoint))
            return
        else:
            end_ip = self.get_endpoint_ip(endpoint)

        try:
            vpp_if = vppif.VppApiInterface('tracehost')
            ippass_tap = vpp_if.get_tap_interface_by_host_ifname(self._br_name)
            if not ippass_tap:
                self.logger.error('Failed to find interface {} for ip-passthrough configuration' .format(self._br_name))
                return

            vpp_if.add_del_ippass_permit_host_entry(ippass_tap['sw_if_index'], 'tcp', ip=end_ip, dst_port=80, add=0 if is_del else 1)
            del (vpp_if)
        except Exception as e:
            self.logger.error('Failed to configure ip-pass rules for {} for interface {} with code {}' .format(endpoint, self._ifname, e.returncode))
            pass

    # ## SAMPLING ##
    def sampling(self):
        if self._ifstate == False or self._trace_state == False:
            return

        for id, pr in self._probes.items():
            count = 0
            upcount = 0
            if pr['type'] == PR_TYPE.ICMP:
                (count, upcount, sended) = self.probeICMP(endpoint=pr['endpoint'], tr_type=pr['tr_type'], max_rtt=(int(pr['max-rtt'])) if 'max-rtt' in pr else 0)

            if pr['type'] == PR_TYPE.THROUGHPUT:
                (count, upcount, sended) = self.probeSpeedtest(endpoint=pr['endpoint'], min_up=pr['min_up'], min_down=pr['min_down'], 
                                                               direction=pr['direction'], cmd=pr['cmd'])
            self.logger.debug("Trace sampling ifname {} probe {} type {} - count {} upcount {}" .format(self._ifname, id, pr['type'], count, upcount))
            pr['pkts_send'] += sended
            pr['pkts_down'] += count
            # self.logger.debug("probe {} statistics - send {} down {}" .format(pr, pr['pkts_send'], pr['pkts_down']))
            
            if self.curr_state == True:
                if count >= self._ch_count:
                    self.need_change = True
                    self.downcount += 1
                    self.probe_to_be_reported = pr
                    raise TracehostErr("Trace-host of interface {} failed" .format(self._ifname), -1)
            else:
                if upcount >= self._ch_count:
                    self.need_change = True
                    self.upcount += 1
                    self.probe_to_be_reported = pr
                    raise TracehostErr("Trace-host of interface {} succeeded" .format(self._ifname), -1)

    def disable_rp_filter(self, ifname, disable=True):
        # current scheme for rp_filter: 
        # config/all/rp_filter=0
        # config/dataX/rp_filter=2
        # config/br-dataX/rp_filter=0

        if not disable:
            return
            
        rp_value = 2 if disable else 1
        if "br-" in ifname and disable:
            rp_value = 0
        FNULL = open(os.devnull, 'w')
        try:
            sysctl_param = 'net.ipv4.conf.%s.rp_filter=%d' % (str(ifname).replace('.', '/'), rp_value)
            subprocess.check_call(['sysctl', sysctl_param], stdout=FNULL, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.logger.error('disable_rp_filter - FAIL')
            self.logger.error('%s returns %d' % (e.cmd, e.returncode))

        FNULL.close()

    # ## PROBES ##
    def probeICMP(self, endpoint, tr_type=None, max_rtt=0):
        self.logger.debug('probeICMP - ifname {} endpoint {} max-rtt' .format(self._ifname, endpoint, max_rtt))
        
        count = 0
        upcount = 0
        i = 0
        self.disable_rp_filter(self._br_name, True)
        FNULL = open(os.devnull, 'w')
        while (i < self._count):
            try:
                i += 1
                if tr_type == PR_ENDP_TYPE.IPV6:
                    if not self._ipv6:
                        self.logger.error('Configuration mismatch - interface {} with probe endpoint {} does not have an IPv6 address' 
                                          .format(self._br_name, endpoint))
                    if int(max_rtt) > 0:
                        output = subprocess.check_output(['nice', '-n-10', 'ping6', '-I', str(self._br_name), str(endpoint), '-c', '1', '-W', str(self._timeout)]).decode()
                        rtt = float((re.search('time=(.*) ms', output)).group(1))
                        if (int(rtt) >= int(max_rtt)):
                            raise RttException('PING to %s via %s rtt %d - exeeded max RTT %d' % (str(endpoint), str(self._br_name), int(rtt), int(max_rtt)))
                    else:
                        subprocess.check_call(['nice', '-n-10', 'ping6', '-I', str(self._br_name), str(endpoint), '-c', '1', '-W', str(self._timeout)],
                                            stdout=FNULL, stderr=subprocess.STDOUT)
                else:
                    if not self._ipv4:
                        self.logger.error('Configuration mismatch - interface {} with probe endpoint {} does not have an IPv4 address' 
                                          .format(self._br_name, endpoint))
                    if int(max_rtt) > 0:
                        output = subprocess.check_output(['nice', '-n-10', 'ping', '-I', str(self._br_name), str(endpoint), '-c', '1', '-W', str(self._timeout)]).decode()
                        rtt = float((re.search('time=(.*) ms', output)).group(1))
                        if (int(rtt) >= int(max_rtt)):
                            raise RttException('PING to %s via %s rtt %d - exeeded max RTT %d' % (str(endpoint), str(self._br_name), int(rtt), int(max_rtt)))
                    else:
                        subprocess.check_call(['nice', '-n-10', 'ping', '-I', str(self._br_name), str(endpoint), '-c', '1', '-W', str(self._timeout)],
                                            stdout=FNULL, stderr=subprocess.STDOUT)
                upcount += 1
            except RttException as rtt_ex:
                self.logger.error(rtt_ex.args[0])
                count += 1
            except:
                self.logger.error('PING to %s via %s - FAIL' % (str(endpoint), str(self._br_name)))
                count += 1

            if (self.curr_state == True and count >= self._ch_count) or \
                    (self.curr_state == False and upcount >= self._ch_count):
                break

        self.disable_rp_filter(self._br_name, False)
        FNULL.close()
        return (count, upcount, i)

    def probeSpeedtest(self, endpoint, min_up, min_down, direction, cmd):
        self.logger.debug('probeSpeedtest - ifname {} endpoint {} dir {} min-up {} min-down {} ' 
            .format(self._ifname, endpoint, direction, min_up, min_down))
        count = 0
        upcount = 0
        i = 0
        while (i < self._count):
            try:
                i += 1
                speedtest_res = subprocess.run(cmd, stdout=subprocess.PIPE).stdout.decode('utf-8')
                res = json.loads(speedtest_res)[0]
                if direction != PR_THR_DIR.DOWNLOAD and float(min_up) != 0 and float(min_up) > float(res['upload']):
                    raise RttException('THR to {} via {} upload {} lower that {}' .format(str(endpoint), str(self._br_name), float(res['upload']), float(min_up)))
                if direction != PR_THR_DIR.UPLOAD and float(min_down) != 0 and float(min_down) > float(res['download']):
                    raise RttException('THR to {} via {} download {} lower that {}' .format(str(endpoint), str(self._br_name), float(res['download']), float(min_down)))
                
                upcount += 1
            except RttException as rtt_ex:
                self.logger.error(rtt_ex.args[0])
                count += 1
            except subprocess.CalledProcessError as e:
                self.logger.error('THR to {} via {} - FAILED' .format(str(endpoint), str(self._br_name)))
                count += 1
            except Exception as e:
                self.logger.error('THR to {} via {} - FAILED' .format(str(endpoint), str(self._br_name)))
                count += 1

            if (self.curr_state == True and count >= self._ch_count) or \
                    (self.curr_state == False and upcount >= self._ch_count):
                break

        return (count, upcount, i)

class TracehostMonitor(object):
    """
    <node>
        <interface name='net.telco.TracehostMonitor'>
            <method name='load_initial'>
                <arg type='i' name='resp' direction='out' />
            </method>
            <method name='load_config'>
                <arg type='s' name='ifname' direction='in' />
                <arg type='i' name='resp' direction='out' />
            </method>
            <method name='del_config'>
                <arg type='s' name='ifname' direction='in' />
                <arg type='i' name='resp' direction='out' />
            </method>
            <method name='get_status'>
                <arg type='s' name='ifname' direction='in' />
                <arg type='i' name='resp' direction='out' />
                <arg type='i' name='status' direction='out' />
            </method>
            <method name='get_statistics'>
                <arg type='s' name='ifname' direction='in' />
                <arg type='i' name='resp' direction='out' />
                <arg type='(uuua(uuuu))' name='trace' direction='out' />
            </method>
        </interface>
    </node>
    """
    
    def __init__(self, loop_obj=None, logger=None, log_level=logging.ERROR, is_vpp=False):
        super(TracehostMonitor,self).__init__()
        self.loop = loop_obj
        self.logger = None
        self.t_monitor = None
        self.monitor_cancelled = True
        self.t_hosts = []
        self.sample_interval = 1
        self.is_vpp = is_vpp
        self.is_nano = True if os.path.isfile("/boot/armbianEnv.txt") else False
        
        if logger:
            self.logger = logger
        else:
            self.__init_logger( log_level )
        
        self.start_monitor()

    def __init_logger(self, log_level ):
        self.logger = logging.getLogger(name=LOG_NAME)
        self.logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self.logger.addHandler(syslog_hnd)
        self.logger.propagate = False

    # ## D-BUS MESSAGES ##

    def load_initial(self):
        resp = 1
        self.logger.info('load_initial()')
        try:
            self.load_initial_config()
            resp = 0
        except TracehostErr as ex:
            self.logger.error('load_initial err - {}' .format(ex))
        return resp
    
    def load_config(self, ifname):
        resp = 1
        srif = sr_if.SysrepoInterfaces(app_name=LOG_NAME)
        # srif.session.refresh()
        self.logger.info('load_config({})'.format(ifname))
        try:
            # ifcfg = next((intf for intf in srif.get_interfaces().keys() if intf == ifname), None)
            # if not ifcfg:
            #     return 1
            inft_cfg = srif.get_interfaces(startup=True)
            for name, ifcfg in inft_cfg.items():
                if name == ifname:
                    self.create_tracehost(name, ifcfg)
                    resp = 0
                    break
        except TracehostErr as ex:
            self.logger.error('load_config err - {}' .format(ex))
        return resp
    
    def del_config(self, ifname):
        resp = 1
        self.logger.info('del_config({})'.format(ifname))
        try:
            self.delete_tracehost(ifname)
            resp = 0
        except TracehostErr as ex:
            self.logger.error('del_config err - {}' .format(ex))
        return resp

    def get_status(self, ifname):
        resp = 1
        status = 0
        self.logger.info('get_status({})'.format(ifname))
        try:
            t_host = self.get_tracehost(ifname)
            if t_host == None:
                raise TracehostErr('Tracehost \'{}\' does not exist' .format(ifname), -1)
            status = t_host.curr_state
            resp = 0
        except TracehostErr as ex:
            self.logger.error('get_status err - {}' .format(ex))
        return (resp, status)
        
    def get_statistics(self, ifname):
        resp = 1
        pr = [(0, 0, 0, 0)]
        trace = (0, 0, 0, pr)
        self.logger.info('get_statistics({})'.format(ifname))
        try:
            t_host = self.get_tracehost(ifname)
            if t_host == None:
                raise TracehostErr('Tracehost \'{}\' does not exist' .format(ifname), -1)
            
            if t_host._probes:
                pr = []
                for id, probe in t_host._probes.items():
                    prr = (id, probe['type'], probe['pkts_send'], probe['pkts_down'])
                    pr.append(prr)
            
            trace = (t_host.curr_state, t_host.upcount, t_host.downcount, pr)
            resp = 0
        except TracehostErr as ex:
            self.logger.error('get_statistics err - {}' .format(ex))
        return (resp, trace)
    # ## END OF D-BUS MESSAGES ##

    def load_initial_config(self):
        srif = sr_if.SysrepoInterfaces(app_name=LOG_NAME)
        # inft_cfg = srif.get_interfaces()
        inft_cfg = srif.get_interfaces(startup=True)
        for ifname, ifcfg in inft_cfg.items():
            iftype = ifcfg.get('type', '')
            if iftype != 'ethernet' and iftype != 'lte' and iftype != 'wifi' and iftype != 'pppoe':
                continue

            # self.logger.debug('load initial ({},{})'.format(ifname, ifcfg))
            self.create_tracehost(ifname, config=ifcfg)

    def create_tracehost(self, ifname, config=None):
        t_host = next(( a for a in self.t_hosts if a.ifname == ifname), None)
        if t_host:
            self.logger.debug("Trace host configuration for interface \'{}\' allready exists.".format(t_host.ifname))
        else:
            tr_config = config.get('tracehost', None)
            if not tr_config:
                return
            if False == tr_config.get('enable', False) or False == config.get('enable', True) or 'probes' not in tr_config:
                return

            pr_list = {}
            for pr in tr_config['probes']:
                if 'endpoint' not in pr or 'type' not in pr:
                    continue
                pr_list[pr['id']] = pr
                pr_list[pr['id']]['type'] = self.get_probe_type(pr.get('type', 'icmp'))
                pr_list[pr['id']]['direction'] = self.get_throughput_dir(pr.get('direction', 'both'))
                pr_list[pr['id']]['min_up'] = round(float(pr.get('min-upload', 0)), 3)
                pr_list[pr['id']]['min_down'] = round(float(pr.get('min-download', 0)), 3)
                if False == self.check_thr(pr_list[pr['id']]['type'], pr_list[pr['id']]['min_up'], pr_list[pr['id']]['min_down'], pr_list[pr['id']]['direction']):
                    return
                pr_list[pr['id']]['pkts_send'] = 0
                pr_list[pr['id']]['pkts_down'] = 0

            self.logger.info("Add trace host configuration for interface \'{}\' - probes {} " .format(ifname, pr_list))
            if_brname = self.get_system_if_name(ifname, config.get('vlan', None))
            t_host = Tracehost(ifname, if_brname=if_brname, logger=self.logger, pr_list=pr_list, 
                                ifstate=config.get('enable', True), trace_state=tr_config.get('enable', False),
                                count=tr_config.get('count', 3), ch_count=tr_config.get('change-count', 3),
                                timeout=tr_config.get('timeout', 3), interval=tr_config.get('interval', 10),
                                metric=config.get('metric', None), state_prop=tr_config.get('trace-state-propagation', False),
                                is_vpp=self.is_vpp, ip_pass=config.get('ip-passthrough', False))

            self.t_hosts.append(t_host)
            t_host.start()

    def get_tracehost(self, ifname):
        if not self.t_hosts:
            return None
        return (next(( a for a in self.t_hosts if a.ifname == ifname), None))

    def delete_tracehost(self, ifname):
        t_host = self.get_tracehost(ifname)
        if t_host == None:
            raise TracehostErr('Tracehost \'{}\' does not exist' .format(ifname), -1)

        t_host.cancel()
        if t_host._ipv4:
            self.trace_modify_route(t_host._br_name, t_host._metric, False, t_host._state_propagation) # cleanup to restore the origin route
        if t_host._ipv6:
            self.trace_modify_route(t_host._br_name, t_host._metric, t_host._ipv6, t_host._state_propagation) # cleanup to restore the origin route
        for id, pr in t_host._probes.items():
            if pr['type'] == PR_TYPE.THROUGHPUT:
                if os.path.exists(pr['srv_file']):
                    os.remove(pr['srv_file'])

        self.t_hosts.remove(t_host)
        del t_host

    def get_probe_type(self, probe_type):
        pr_type = PR_TYPE.ICMP
        
        if probe_type == 'throughput':
            pr_type = PR_TYPE.THROUGHPUT
        elif probe_type == 'tcp':
            pr_type = PR_TYPE.TCP
        elif probe_type == 'http':
            pr_type = PR_TYPE.HTTP
        elif probe_type == 'https':
            pr_type = PR_TYPE.HTTPS
        elif probe_type == 'arp':
            pr_type = PR_TYPE.ARP
        elif probe_type == 'dns':
            pr_type = PR_TYPE.DNS
        
        return pr_type

    def get_throughput_dir(self, probe_dir):
        pr_dir = PR_THR_DIR.BOTH
        
        if probe_dir == 'upload':
            pr_dir = PR_THR_DIR.UPLOAD
        elif probe_dir == 'download':
            pr_dir = PR_THR_DIR.DOWNLOAD
        
        return pr_dir

    def check_thr(self, pr_type, min_up, min_down, pr_dir):
        if pr_type != PR_TYPE.THROUGHPUT:
            return True
        if min_up == 0 and (pr_dir == PR_THR_DIR.BOTH or pr_dir == PR_THR_DIR.UPLOAD):
            return False
        if min_down == 0 and (pr_dir == PR_THR_DIR.BOTH or pr_dir == PR_THR_DIR.DOWNLOAD):
            return False
        return True
    # ## TRACE MONITOR FUNC ##
    
    def start_monitor(self):
        if self.t_monitor:
            self.logger.info('trace monitoring already running')
            return
        self.logger.info('start trace monitoring if states({},{})'.format(self.t_monitor, self.t_hosts))
        self.monitor_cancelled = False
        self.t_monitor = Thread(target=self.trace_monitor_th)
        self.t_monitor.start()
        
    def stop_monitor(self):
        if not self.t_monitor:
            self.logger.info('monitor does not exists, return')
            return
        self.logger.info('stop trace monitoring if states({},{})'.format(self.t_monitor, self.t_hosts))
        self.monitor_cancelled = True
    
    def trace_monitor_th(self):
        global monitor_loop
        while not self.monitor_cancelled:
            try:
                if self.monitor_cancelled:
                    break
                if not monitor_loop:
                    break
                for t in self.t_hosts:
                    if t.need_change == True:
                        self.logger.debug('trace_monitor_th ifname {} - curr {} ' .format(t.ifname, t.curr_state))
                        if t._ipv6:
                            self.trace_modify_route(t._br_name, t._metric if t.curr_state == False else 1024, t._ipv6, t._state_propagation, worst=0)
                        if t._ipv4:
                            self.trace_modify_route(t._br_name, t._metric if t.curr_state == False else 255, False, t._state_propagation, worst=0)
                        t.need_change = False
                        t.curr_state = not t.curr_state
                        self.send_notification(t)
                        t.probe_to_be_reported = None
                        self.logger.debug('trace_monitor_th ifname {} -CH TO {} ' .format(t.ifname, t.curr_state))
            except TracehostErr as e:
                self.logger.error('trace monitoring exception - {}' .format(e))
                break
            except Exception as e:
                self.logger.error('trace monitoring exception - {}' .format(e))
                break
            
            time.sleep(self.sample_interval)
        
        self.t_monitor = None
        self.logger.info('Stop trace monitoring')
        self.logger.info('stop all probes')
        for t in self.t_hosts:
            t.cancel()
            if t._ipv4:
                self.trace_modify_route(t._br_name, t._metric, False, t._state_propagation) # cleanup to restore the origin route
            if t._ipv6:
                self.trace_modify_route(t._br_name, t._metric, t._ipv6, t._state_propagation) # cleanup to restore the origin route
        
        self.t_hosts = []
            
        # sys.exit(0)

    # ## END OF TRACE MONITOR FUNC ##

    def get_interface_vtag(self, ifvlan):
        if self.is_vpp == True:
            return 0

        if 'int' == type(ifvlan).__name__:
            return ifvlan

        if ifvlan is None or ifvlan == '': return 0

        vtag = int(ifvlan)

        return vtag

    def get_system_if_name(self, interface, ifvlan):
        vlan = self.get_interface_vtag(ifvlan)
        if 0 == vlan or 'inband.' in interface:
            ifname = interface
        else:
            ifname = "%s.%d" % (interface, vlan)

        if 'data' in ifname or 'lte' in ifname or 'wifi' in ifname:
            ifname = "br-%s" % ifname

        return ifname

    def check_default_metric(self, metric, dev_name):
        output = subprocess.check_output(['ip', 'route', 'list']).decode()
        for ln in output.splitlines():
            if 'default' in ln:
                if 'dev ' + dev_name in ln:
                    continue
                count = 0
                fields = ln.split(' ')
                for f in fields:
                    if 'metric' in f:
                        if int(fields[count + 1]) == int(metric):
                            return 0
                        else:
                            break

                    count += 1
        return 1

    def get_def_gw(self):
        min_metric = None
        ifname = None
        iproutes = subprocess.check_output(['ip', 'route', 'list', 'exact', '0.0.0.0/0']).decode()

        for route in iproutes.splitlines():
            route = route.replace(" proto unspec", "")
            rt_line = route.split(' ')
            rt_line = [x for x in rt_line if x != '']
            # default via 192.168.8.1 dev lte0 metric 100
            if not min_metric:
                min_metric = int(rt_line[6])
                ifname = str(rt_line[4])
            elif int(rt_line[6]) < min_metric:
                min_metric = int(rt_line[6])
                ifname = str(rt_line[4])

        return ifname

    def ovpn_client_restart(self):
        try:
            self.logger.info('Restarting OpenVPN client...')
            subprocess.check_call(['systemctl', 'restart', 'openvpn@client.service'])
        except subprocess.CalledProcessError as e:
            self.logger.error('Command \'%s\' failed with code%d' % (e.cmd, e.returncode))
        except OSError as e:
            self.logger.error('System error - \'%s\'' % e.strerror)

    def ovpn_check_gw(self):
        if not os.path.exists(OVPN_DEF_GW):
            return

        ovpn_file = open(OVPN_DEF_GW, 'r')
        ovpn_gw = ovpn_file.read()
        ovpn_file.close()
        if not ovpn_gw:
            self.logger.error('Failed to get ovpn config !')
            return

        ifname = self.get_def_gw()
        if not ifname:
            self.logger.error('Failed to get ip routes !')
            return

        if ovpn_gw != ifname:
            self.ovpn_client_restart()

    def nano_state_propagation(self, ifname, state):
        clav_if = ''
        clav_path=None

        if os.path.exists('/root/apps/clavister/clav_ifs_setup'):
            clav_path = '/root/apps/clavister/clav_ifs_setup'
        elif os.path.exists('/root/apps/clavistervnf/clav_ifs_setup'):
            clav_path = '/root/apps/clavistervnf/clav_ifs_setup'
        
        if clav_path:
            with open(clav_path) as fl:
                for line in fl:
                    if str(ifname) in line:
                        clav_if = str(line.split()[-1])
                fl.close()

        if clav_if:
            FNULL = open(os.devnull, 'w')
            subprocess.check_call(['ip', 'link', 'set', 'dev', str(clav_if), str(state)],
                                            stdout=FNULL, stderr=subprocess.STDOUT)
            FNULL.close()
            self.logger.info('Change interface {} state to {}' .format(clav_if, state))

    def trace_modify_route(self, ifname, metric, ipv6, prop_vnics, worst=0):
        if self.is_vpp == True:
            UPDOWNTOOL = '/scripts/eth_up_down_vpplauncher.sh'
        else:
            UPDOWNTOOL = '/scripts/eth_up_down_launcher.sh'

        if 255 == metric or 1024 == metric:
            while (metric > 0):
                if 1 == self.check_default_metric(metric, ifname):
                    break
                metric -= 1
        if ipv6:
            output = subprocess.check_output(['ip', '-6', 'route', 'list']).decode()
            worst = 1024
        else:
            output = subprocess.check_output(['ip', 'route', 'list']).decode()
            worst = 255
        for ln in output.splitlines():
            if ifname in str(ln) and 'default' in str(ln):
                fields = str(ln).split(' ')
                count = 0
                current_metric = 0
                for f in fields:
                    if 'via' in f:
                        nexthop = fields[count + 1]
                    if 'metric' in f:
                        current_metric = int(fields[count + 1])

                    count += 1

                if metric != current_metric and current_metric <= worst:
                    if ('default via %s dev %s metric %s' % (nexthop, ifname, metric)) not in output:
                        self.logger.info('Going to change metric of %s from %s to %s,nexthop %s' % (
                            ifname, current_metric, metric, nexthop))
                        try:
                            if ipv6:
                                subprocess.check_call(['ip', '-6', 'route', 'del', 'default', 'dev', str(ifname)])
                                defgw.del_def_gw(str(ifname), str(nexthop))
                                subprocess.check_call(
                                    ['ip', '-6', 'route', 'add', 'default', 'via', str(nexthop), 'dev', str(ifname),
                                    'metric', str(metric)])
                                defgw.set_def_gw(str(ifname), str(nexthop), metric)
                            else:
                                subprocess.check_call(['ip', 'route', 'del', 'default', 'dev', str(ifname)])
                                defgw.del_def_gw(str(ifname), str(nexthop))
                                subprocess.check_call(
                                    ['ip', 'route', 'add', 'default', 'via', str(nexthop), 'dev', str(ifname), 'metric',
                                    str(metric)])
                                defgw.set_def_gw(str(ifname), str(nexthop), metric)
                            if prop_vnics == True or prop_vnics == 'true' or prop_vnics == 'True':
                                if current_metric <= metric:
                                    if self.is_nano:
                                        self.nano_state_propagation(str(ifname).strip('br-'), 'down')
                                    else:
                                        subprocess.run([UPDOWNTOOL, str(ifname).strip('br-'), 'DOWN'])
                                else:
                                    if self.is_nano:
                                        self.nano_state_propagation(str(ifname).strip('br-'), 'up')
                                    else:
                                        subprocess.run([UPDOWNTOOL, str(ifname).strip('br-'), 'UP'])
                        except subprocess.CalledProcessError as e:
                            self.logger.error('Command \'%s\' failed with code%d' % (e.cmd, e.returncode))
                        except OSError as e:
                            self.logger.error('System error - \'%s\'' % e.strerror)

        self.ovpn_check_gw()

    def send_notification(self, t_host):
        probe = t_host.probe_to_be_reported
        probe_id = probe.get("id", '')
        type = PR_TYPE(probe.get("type", 0)).name if probe.get("type", 0) else ''
        endpoint = probe.get("endpoint", '')
        message = f'Trace-host on interface \'{t_host.ifname}\' {"succeeded" if t_host.curr_state else "failed"} - probe id \'{probe_id}\' type \'{type}\' endpoint \'{endpoint}\'. '
        severity = "okay" if t_host.curr_state else "failure"
        plugin = "tracehost"

        tm = time.time()
        
        try:
            with open('/config/collectd-hostname', 'r') as f:
                host = f.readline().strip()
        except:
            host=socket.gethostname()

        msg = f'PUTNOTIF plugin={plugin} plugin_instance={t_host.ifname} host={host} severity={severity} time={tm} message="{message}"\n'

        if os.path.exists(COLLECTD_PATH):
            ch = []
            cl = None
            try:
                cl = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                cl.connect(COLLECTD_PATH)
                cl.sendall(msg.encode('utf-8'))
                cl.setblocking(0)
                ch = cl.recv(1)
                while ch != "\n":
                    ch = cl.recv(1)
            except Exception as ex:
                pass
            finally:
                if cl:
                    cl.close()


def print_help(name):
    print(name)

def sig_handler(signal, fname):
    global monitor_loop
    
    monitor_loop = False
    # sys.exit(0)

def daemonize():
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent
            #print "Parent exit"
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

    # redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    null=os.open('/dev/null', os.O_RDWR)
    os.dup2(null, sys.stdin.fileno())
    os.dup2(null, sys.stdout.fileno())
    os.dup2(null, sys.stderr.fileno())

def run_dbus(log_level):
    global loop
    
    loop = GLib.MainLoop()
    bus = SystemBus()
    srif = sr_if.SysrepoInterfaces(app_name=LOG_NAME)
    is_vpp = True if srif.get_network_mode() == 'vpp' else False
    trace_host_m = TracehostMonitor(loop_obj=loop, logger=None, log_level=log_level, is_vpp=is_vpp)

    bus.publish("net.telco.TracehostMonitor", trace_host_m )

    loop.run()

def main(name, argv):
    global monitor_loop
    monitor_loop = True
    daemon = False
    log_level = logging.ERROR
    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGINT, sig_handler)

    try:
        opts, args = getopt.getopt(argv,"dhD",["daemon", "help", "debug" ])
    except getopt.GetoptError as err:
        print(err)
        print_help( name )
        sys.exit(2)

    for opt, arg in opts:
        if opt in ('-h', '--help'):
            print_help( name )
            sys.exit(0)
        elif opt in ("-d", "--daemon"):
            daemon = True
        elif opt in ("-D", "--debug"):
            log_level = logging.DEBUG
        else:
            print("Unknown option \'{}\' !!!".format(opt))
            print_help( name )
            sys.exit(2)

    if daemon:
        daemonize()
        
    run_dbus(log_level)
    

if __name__ == "__main__":
    from scapy import all as sc
    from scapy.data import (
            ARPHDR_ETHER,
            ARPHDR_LOOPBACK,
            ARPHDR_PPP,
            ARPHDR_TUN,
            IPV6_ADDR_GLOBAL
        )
    from sysrepo_interfaces import sysrepo_interfaces3 as sr_if
    import vppapi_if.vppapi_if as vppif
    import default_gw_set as defgw
    from pyroute2 import IPDB

    main(sys.argv[0], sys.argv[1:])
